# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from contextlib import closing
import json
import keyring
from keyring.backends.chainer import ChainerBackend
from keyring.credentials import Credential
import logging
import os

from komikku.utils import get_data_dir

keyring.core.init_backend()

logger = logging.getLogger('komikku')


class CustomCredential(Credential):
    """Custom credentials implementation with an additional 'address' attribute"""

    def __init__(self, username, password, address=None):
        self._username = username
        self._password = password
        self._address = address

    @property
    def address(self):
        return self._address

    @property
    def password(self):
        return self._password

    @property
    def username(self):
        return self._username


class KeyringHelper:
    """Simple helper to store servers accounts credentials using Python keyring library"""

    appid = 'info.febvre.Komikku'

    def __init__(self, fallback_keyring='plaintext'):
        if self.is_disabled or not self.has_recommended_backend:
            if fallback_keyring == 'plaintext':
                keyring.set_keyring(PlaintextKeyring())

    @property
    def has_recommended_backend(self):
        """ Returns True if at least one supported backend is available, False otherwise """

        # At this time, SecretService is the only backend that support a collection of items with arbitrary attributes
        # Known working implementations are:
        # - GNOME Keyring
        # - KeePassXC Secret Service integration (tested, work well)

        # Freedesktop.org Secret Service specification was written by both GNOME and KDE projects together
        # but it’s supported by the GNOME Keyring only
        # ksecretservice (https://community.kde.org/KDE_Utils/ksecretsservice) exists but is unfinished and seems unmaintained
        current_keyring = self.keyring
        return current_keyring is not None and isinstance(current_keyring, keyring.backends.SecretService.Keyring)

    @property
    def is_disabled(self):
        return hasattr(keyring.backends, 'null') and isinstance(self.keyring, keyring.backends.null.Keyring)

    @property
    def keyring(self):
        current_keyring = keyring.get_keyring()

        if isinstance(current_keyring, ChainerBackend):
            # Search SecretService backend
            for backend in current_keyring.backends:
                if isinstance(backend, keyring.backends.SecretService.Keyring):
                    return backend

            return None

        return current_keyring

    def get(self, service):
        if self.is_disabled:
            return None

        current_keyring = self.keyring
        if isinstance(current_keyring, keyring.backends.SecretService.Keyring):
            collection = current_keyring.get_preferred_collection()

            credential = None
            with closing(collection.connection):
                items = collection.search_items({'service': service})
                for item in items:
                    current_keyring.unlock(item)
                    username = item.get_attributes().get('username')
                    if username is None:
                        # Try to find username in 'login' attribute instead of 'username'
                        # Backward compatibility with the previous implementation which used libsecret
                        username = item.get_attributes().get('login')
                    if username:
                        credential = CustomCredential(username, item.get_secret().decode('utf-8'), item.get_attributes().get('address'))
        else:
            # Fallback backend
            credential = current_keyring.get_credential(service, None)

        if credential is None or credential.username is None:
            return None

        return credential

    def store(self, service, username, password, address=None):
        if self.is_disabled:
            return

        current_keyring = self.keyring
        if isinstance(current_keyring, keyring.backends.SecretService.Keyring):
            collection = current_keyring.get_preferred_collection()

            label = f'{self.appid}: {username}@{service}'
            attributes = {
                'application': self.appid,
                'service': service,
                'username': username,
            }
            if address is not None:
                attributes['address'] = address

            with closing(collection.connection):
                # Delete previous credential if exists
                items = collection.search_items({'service': service})
                for item in items:
                    item.delete()

                collection.create_item(label, attributes, password)
        else:
            # Fallback backend
            current_keyring.set_password(service, username, password, address)


class PlaintextKeyring(keyring.backend.KeyringBackend):
    """Simple File Keyring with no encryption

    Used as fallback when no supported Keyring backend is found
    """

    priority = 1

    def __init__(self):
        super().__init__()

    @property
    def filename(self):
        return os.path.join(self.folder, 'plaintext.keyring')

    @property
    def folder(self):
        return os.path.join(get_data_dir(), 'keyrings')

    def _read(self):
        if not os.path.exists(self.filename):
            return {}

        with open(self.filename, 'r') as fp:
            return json.load(fp)

    def _save(self, data):
        if not os.path.exists(self.folder):
            os.mkdir(self.folder)

        with open(self.filename, 'w+') as fp:
            json.dump(data, fp, indent=2)

        os.chmod(self.filename, 0o600)

    def get_credential(self, service, _username):
        data = self._read()
        if service in data:
            return CustomCredential(data[service]['username'], data[service]['password'], data[service].get('address'))
        return None

    def get_password(self, service, username):
        pass

    def set_password(self, service, username, password, address=None):
        data = self._read()
        data[service] = dict(
            username=username,
            password=password,
        )
        if address is not None:
            data[service]['address'] = address
        self._save(data)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import textwrap

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import TextImage
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent


class Grisebouille(Server):
    id = 'grisebouille'
    name = 'Grise Bouille'
    lang = 'fr'
    content = ServerContent(
        type=[_('Webcomic'), _('Self-publishing')],
        license='CC BY-SA'
    )
    true_search = False

    long_strip_genres = ['Long Strip', ]

    base_url = 'https://grisebouille.net'
    donate_url = 'https://soutenir.ptilouk.net/'
    logo_url = base_url + '/content/img/favicon-32x32.png'
    search_url = base_url + '/series/'
    manga_url = base_url + '/serie/{0}/'
    chapter_url = base_url + '/{0}/'
    cover_url = base_url + '/content/img/logo-{0}.png'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content
        """
        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=['Gee', ],
            scanlators=[],
            genres=['Humour', 'Long Strip'],
            status='ongoing',
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=self.cover_url,
        ))

        data['name'] = soup.select_one('title').text.split('|')[0].strip().encode('iso-8859-1').decode()
        data['cover'] = self.cover_url.format(data['slug'])

        data['synopsis'] = soup.select_one('.main.block p').text.strip().encode('iso-8859-1').decode()

        # Chapters
        for a_element in reversed(soup.select('.episodes a')):
            slug = a_element.get('href').split('/')[-1]
            if data['slug'] == 'superflu' and not slug.startswith('s0'):
                continue

            title = a_element.select_one('.episode-name').text.strip().encode('iso-8859-1').decode()
            num = a_element.select_one('.episode-nb').text.strip()[1:-1]
            date = a_element.select_one('.episode-date').text.strip()

            data['chapters'].append(dict(
                slug=slug,
                title=f'#{num} {title}',
                date=convert_date_string(date, '%Y-%m-%d'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for index, element in enumerate(soup.select('.main.block > p, .main.block :has(> img)')):
            if img_element := element.img:
                url = img_element.get('src')
                if not url.startswith(self.base_url):
                    continue

                data['pages'].append(dict(
                    slug=None,
                    image=url.encode('iso-8859-1').decode(),
                ))
            else:
                text = element.text.strip()
                try:
                    text = text.encode('iso-8859-1').decode()
                except Exception:
                    pass
                data['pages'].append(dict(
                    slug=None,
                    image=None,
                    text=text,
                    index=index + 1,
                ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page (image or text) content
        """
        if page.get('image'):
            r = self.session_get(page['image'])
            if r.status_code != 200:
                return None

            mime_type = get_buffer_mime_type(r.content)
            if not mime_type.startswith('image'):
                return None

            name = page['image'].split('/')[-1]
            content = r.content
        else:
            text = '\n'.join(textwrap.wrap(page['text'], 25))
            image = TextImage(text)

            mime_type = image.mime_type
            name = f'txt_{page["index"]:03d}.{image.format}'  # noqa: E231
            content = image.content

        return dict(
            buffer=content,
            mime_type=mime_type,
            name=name,
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        r = self.session_get(self.search_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = []
        for a_element in soup.select('.series-link'):
            slug = a_element.get('href').split('/')[-1]
            if slug not in ('comic-trip', 'depeches-melba', 'la-chaine-meteore', 'la-fourche', 'superflu', 'tu-sais-quoi'):
                continue

            data.append(dict(
                slug=slug,
                name=a_element.img.get('alt').encode('iso-8859-1').decode(),
                cover=a_element.img.get('src'),
            ))

        return data

    def search(self, term=None):
        # This server does not have a search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.foolslide import FoOlSlide


class Lecercleduscan(FoOlSlide):
    id = 'lecercleduscan'
    name = 'Le Cercle du Scan'
    lang = 'fr'
    status = 'disabled'

    base_url = 'https://lel.lecercleduscan.com'
    search_url = base_url + '/search/'
    mangas_url = base_url + '/directory/'
    manga_url = base_url + '/series/{0}'
    chapter_url = base_url + '/read/{0}/fr/{1}/page/1'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# BEWARE: Old Hatigarmscans is disabled
# Dead since 02/2020
# The site has changed its address and is now available in the genkan server

from komikku.servers.multi.genkan import GenkanInitial
from komikku.servers.multi.my_manga_reader_cms import MyMangaReaderCMS


class Hatigarmscans(GenkanInitial):
    id = 'hatigarmscans'
    name = 'Hatigarm Scans'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://hatigarmscanz.net'
    search_url = base_url + '/comics'
    most_populars_url = base_url + '/home'
    manga_url = base_url + '/comics/{0}'
    chapter_url = base_url + '/comics/{0}/{1}'
    image_url = base_url + '{0}'


class Hatigarmscans__old(MyMangaReaderCMS):
    id = 'hatigarmscans__old'
    name = 'Hatigarm Scans'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://www.hatigarmscans.net'
    search_url = base_url + '/search'
    most_populars_url = base_url + '/filterList?page=1&sortBy=views&asc=false'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}'
    image_url = base_url + '/uploads/manga/{0}/chapters/{1}/{2}'
    cover_url = base_url + '/uploads/manga/{0}/cover/cover_250x350.jpg'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.peachscan import Peachscan


class Rfdragonscan(Peachscan):
    id = 'rfdragonscan'
    name = 'RF Dragon Scan'
    lang = 'pt_BR'
    status = 'disabled'

    has_cf = True

    base_url = 'https://rfdragonscan.net'



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
from io import BytesIO
import threading

from PIL import Image
try:
    import pytesseract
except Exception:
    pytesseract = None

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import GtkSource

from komikku.models.settings import Settings
from komikku.translators import Google
from komikku.translators import LANGUAGES


class LangCodeNamePair(GObject.Object):
    code = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE, default='')
    name = GObject.Property(type=str, nick='Name', blurb='Name', flags=GObject.ParamFlags.READWRITE, default='')


class OCRTranslator(GObject.GObject):
    def __init__(self, reader):
        super().__init__()

        self.__active = False
        self.enabled = True

        self.reader = reader
        self.window = reader.window

        self.togglebutton = self.reader.ocr_translator_togglebutton
        self.bottomsheet = self.reader.ocr_translator_bottomsheet
        self.box = self.reader.ocr_translator_box
        self.src_dropdown = self.reader.ocr_translator_src_dropdown
        self.src_scrolledwindow = self.reader.ocr_translator_src_scrolledwindow
        self.dst_dropdown = self.reader.ocr_translator_dst_dropdown
        self.dst_scrolledwindow = self.reader.ocr_translator_dst_scrolledwindow
        self.clear_button = self.reader.ocr_translator_clear_button
        self.chars_counter_label = self.reader.ocr_translator_chars_counter_label
        self.translate_button = self.reader.ocr_translator_translate_button
        self.translate_spinner = self.reader.ocr_translator_translate_spinner
        self.copy_button = self.reader.ocr_translator_copy_button

        if pytesseract is None or not self.ocr_languages:
            self.enabled = False
            self.togglebutton.set_visible(False)
        else:
            self.reader.window.breakpoint.add_setter(self.box, 'orientation', Gtk.Orientation.VERTICAL)

            self.togglebutton.connect('toggled', self.toggle)

            list_store_expression = Gtk.PropertyExpression.new(LangCodeNamePair, None, 'name')
            items = [LangCodeNamePair(code='auto', name='Auto')]
            items += [LangCodeNamePair(code=k, name=v) for k, v in LANGUAGES.items()]

            # Source
            model = Gio.ListStore(item_type=LangCodeNamePair)
            model.splice(0, 0, items)
            self.src_dropdown.set_expression(list_store_expression)
            self.src_dropdown.set_model(model)

            self.src_text = TextView(editable=True, bottom_margin=9, left_margin=9, right_margin=9, top_margin=9)
            self.src_text.buffer.connect('changed', self.on_src_text_changed)
            self.src_scrolledwindow.set_child(self.src_text)

            # Destination
            model = Gio.ListStore(item_type=LangCodeNamePair)
            model.splice(0, 0, items[1:])
            self.dst_dropdown.set_expression(list_store_expression)
            self.dst_dropdown.set_model(model)
            self.dst_dropdown.set_selected(list(LANGUAGES.keys()).index('en'))

            self.dst_text = TextView(editable=False, bottom_margin=9, left_margin=9, right_margin=9, top_margin=9)
            self.dst_scrolledwindow.set_child(self.dst_text)

            # Buttons
            self.clear_button.connect('clicked', self.src_text.clear_text)
            self.translate_button.connect('clicked', self.translate)
            self.copy_button.connect('clicked', self.copy_dst)

    @GObject.Property(type=bool, default=False)
    def active(self):
        return self.__active

    @active.setter
    def active(self, active):
        if not self.enabled:
            return
        if active and self.reader.window.page != self.reader.props.tag:
            # Avoid activation outside of reader page
            return

        self.__active = active

        self.reader.pager.interactive = not active
        self.bottomsheet.props.reveal_bottom_bar = active
        self.togglebutton.props.active = active
        if not active:
            self.bottomsheet.props.open = False
            self.togglebutton.remove_css_class('accent')
        else:
            self.togglebutton.add_css_class('accent')

    @property
    def ocr_lang(self):
        lang = self.reader.manga.ocr_lang
        if lang and lang in self.ocr_languages:
            return lang

        lang = Settings.get_default().ocr_lang
        if lang and lang in self.ocr_languages:
            return lang

        return self.ocr_languages[0]

    @property
    def ocr_languages(self):
        if pytesseract is None:
            return None

        languages = pytesseract.get_languages()
        for lang in ('osd', 'equ'):
            if lang in languages:
                languages.remove(lang)

        return languages

    def copy_dst(self, _btn):
        if display := Gdk.Display.get_default():
            display.get_clipboard().set(self.dst_text.get_text())
            self.reader.window.add_notification(_('Copied to clipboard'))

    def on_src_text_changed(self, _buffer):
        text = self.src_text.get_text()
        self.chars_counter_label.set_text(f'{len(text)}/2000')
        self.translate_button.set_sensitive(text and not text.isspace())

    def open_sheet(self, text):
        self.dst_text.clear_text()
        if text:
            self.src_text.set_text(text)
            self.translate_button.set_sensitive(True)
            self.bottomsheet.props.open = True

    def recognize(self, image, x, y, w, h):
        def complete(text):
            self.open_sheet(text)

        def run():
            with Image.open(image.path or BytesIO(image.data)) as full_img:
                img = full_img.crop((x, y, w, h))
                text = pytesseract.image_to_string(img, lang=self.ocr_lang, config='--psm 12 --oem 1')
                img.close()

            text = text.strip().replace('\n\n', '\n')

            GLib.idle_add(complete, text)

        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()

    def set_active(self, active):
        self.active = active

    def toggle(self, _button):
        active = self.togglebutton.get_active()
        if active == self.active:
            return

        self.active = active

    def translate(self, _btn):
        def on_complete(result):
            self.translate_spinner.set_visible(False)
            self.dst_text.set_text(result['translated'])

        def on_error(message):
            self.translate_spinner.set_visible(False)
            self.window.add_notification(message)

        def run(text):
            try:
                result = Google().translate(
                    text,
                    src=self.src_dropdown.get_selected_item().code,
                    dst=self.dst_dropdown.get_selected_item().code
                )
            except Exception as e:
                GLib.idle_add(on_error, e.message)
                return

            GLib.idle_add(on_complete, result)

        text = self.src_text.get_text().strip()
        self.translate_spinner.set_visible(True)
        self.dst_text.clear_text()

        thread = threading.Thread(target=run, args=(text, ))
        thread.daemon = True
        thread.start()


class TextView(GtkSource.View):
    __gtype_name__ = 'TextView'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.buffer = self.props.buffer
        scheme = 'adwaita-dark' if Adw.StyleManager.get_default().get_dark() else 'adwaita-light'
        self.buffer.set_style_scheme(GtkSource.StyleSchemeManager.get_default().get_scheme(scheme))

    def clear_text(self, *args):
        self.set_text('')

    def get_text(self):
        return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True)

    def set_text(self, text):
        self.buffer.set_text(text)



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _

BORDERS_CROP_THRESHOLD_MIN = 100
BORDERS_CROP_THRESHOLD_MAX = 240
BORDERS_CROP_THRESHOLD_STEP = 10
BORDERS_CROP_THRESHOLDS = range(
    BORDERS_CROP_THRESHOLD_MIN,
    BORDERS_CROP_THRESHOLD_MAX + BORDERS_CROP_THRESHOLD_STEP,
    BORDERS_CROP_THRESHOLD_STEP
)
COVER_WIDTH = 180
COVER_HEIGHT = 256
LOGO_SIZE = 32
MISSING_IMG_RESOURCE_PATH = '/info/febvre/Komikku/images/missing_file.png'
PROGRESSBAR_THEMES = {
    'accent-color': {
        'name': _('Accent Color'),
        'colors': None,
    },
    # https://en.wikipedia.org/wiki/Pride_flag
    'pride-rainbow': {
        'name': _('Pride Colors'),
        'colors': ['#E40303', '#FF8C00', '#FFED00', '#008026', '#24408E', '#732982'],
    },
    'lesbian-pride': {
        'name': _('Lesbian Pride Colors'),
        'colors': ['#D62800', '#EF7627', '#FF9B56', '#FFFFFF', '#D162A4', '#B55690', '#A30262'],
    },
    'gay-pride': {
        'name': _('Male Homosexual Pride Colors'),
        'colors': ['#018E71', '#21CFAC', '#9AE9C3', '#FFFFFF', '#7CAFE4', '#4F47CC', '#3C1379'],
    },
    'transgender': {
        'name': _('Transgender Pride Colors'),
        'colors': ['#5BCEFA', '#F5A9B8', '#FFFFFF', '#F5A9B8', '#5BCEFA'],
    },
    'nonbinary': {
        'name': _('Nonbinary Pride Colors'),
        'colors': ['#FCF434', '#FFFFFF', '#9C59D1', '#2C2C2C'],
    },
    'bisexual': {
        'name': _('Bisexual Pride Colors'),
        'colors': ['#D60270', '#D60270', '#9B4F96', '#0038A8', '#0038A8'],
    },
    'asexual': {
        'name': _('Asexual Pride Colors'),
        'colors': ['#000000', '#A3A3A3', '#FFFFFF', '#810081'],
    },
    'pansexual': {
        'name': _('Pansexual Pride Colors'),
        'colors': ['#FF218C', '#FFD800', '#21B1FF'],
    },
    'aromantic': {
        'name': _('Aromantic Pride Colors'),
        'colors': ['#3DA542', '#A7D379', '#FFFFFF', '#A9A9A9', '#000000'],
    },
    'genderfluid': {
        'name': _('Genderfluid Pride Colors'),
        'colors': ['#FF76A4', '#FFFFFF', '#C011D7', '#000000', '#2F3CBE'],
    },
    'polysexual': {
        'name': _('Polysexual Pride Colors'),
        'colors': ['#F61CB9', '#07D569', '#1C92F6'],
    },
    'omnisexual': {
        'name': _('Omnisexual Pride Colors'),
        'colors': ['#FF9CCE', '#FF52BF', '#200044', '#675FFF', '#8DA7FF'],
    },
    'aroace': {
        'name': _('Aroace Pride Colors'),
        'colors': ['#E28C00', '#ECCD00', '#FFFFFF', '#62AEDC', '#203856'],
    },
    'agender': {
        'name': _('Agender Pride Colors'),
        'colors': ['#000000', '#BCC4C7', '#FFFFFF', '#B7F684', '#FFFFFF', '#BCC4C7', '#000000'],
    },
    'genderqueer': {
        'name': _('Genderqueer Pride Colors'),
        'colors': ['#B57EDC', '#FFFFFF', '#4A8123'],
    },
    'intersex': {
        'name': _('Intersex Pride Colors'),
        'colors': ['#FFD800', '#FFD800', '#7902AA', '#FFD800', '#FFD800'],
    },
    'demigender': {
        'name': _('Demigender Pride Colors'),
        'colors': ['#7F7F7F', '#C3C3C3', '#FBFF74', '#FFFFFF', '#FBFF74', '#C3C3C3', '#7F7F7F'],
    },
    'biromantic': {
        'name': _('Biromantic Pride Colors'),
        'colors': ['#8869A5', '#D8A7D8', '#FFFFFF', '#FDB18D', '#151638'],
    },
    'disability': {
        'name': _('Disability Pride Colors'),
        'colors': ['#595959', '#CF7280', '#EEDE77', '#E8E8E8', '#7BC2E0', '#3BB07D', '#595959'],
    },
    'femboy': {
        'name': _('Femboy Pride Colors'),
        'colors': ['#D460A7', '#E4ADCD', '#FFFFFF', '#57CEF8', '#FFFFFF', '#E4ADCD', '#D460A7'],
    },
    'neutrois': {
        'name': _('Neutrois Pride Colors'),
        'colors': ['#FFFFFF', '#1F9F00', '#000000'],
    },
    'random': {
        'name': _('Random Colors'),
        'lenght': 16,
        'colors': None,
    },
}

DOWNLOAD_MAX_DELAY = 1  # in seconds
REQUESTS_TIMEOUT = 5
USER_AGENT = 'Mozilla/5.0 (X11; Linux x86_64; rv:129.0) Gecko/20100101 Firefox/129.0'
USER_AGENT_MOBILE = 'Mozilla/5.0 (Linux; U; Android 4.1.1; en-gb; Build/KLP) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30'

TIMEZONE = datetime.datetime.now(tz=datetime.UTC).astimezone().tzinfo

CREDITS = dict(
    artists=(
        'Tobias Bernard (bertob)',
    ),
    designers=(
        'Tobias Bernard (bertob)',
        'Valéry Febvre (valos)',
    ),
    developers=(
        'Mufeed Ali (fushinari)',
        'Gerben Droogers (Tijder)',
        'Valéry Febvre (valos)',
        'Aurélien Hamy (aunetx)',
        'Amelia Joison (amnetrine)',
        'David Keller (BlobCodes)',
        'Oleg Kiryazov (CakesTwix)',
        'Lili Kurek',
        'Liliana Prikler',
        'Sabri Ünal',
        'Romain Vaudois',
        'Arthur Williams (TAAPArthur)',
        'GrownNed',
        'ISO-morphism',
        'jaskaranSM',
    ),
    translators=(
        'abidin toumi (Arabic)',
        'Rayen Ghanmi (Arabic)',
        'Mohamed Abdalah Noh (Arabic)',
        'Ahmed Najmawi (Arabic)',
        'Jsus “Jsux” (Bengali)',
        'Bhalet Chakma (Bengali)',
        'Rafael Fontenelle (Brazilian Portuguese)',
        'Infinitive Witch (Brazilian Portuguese)',
        'Unidealistic Raccoon (Brazilian Portuguese)',
        'Alex Carvalho (Brazilian Portuguese)',
        'Juliano de Souza Camargo (Brazilian Portuguese)',
        'Giovanne Menicheli (Brazilian Portuguese)',
        'Fúlvio Alves (Brazilian Portuguese)',
        'Felipe (Brazilian Portuguese)',
        'Matheus Santana (Brazilian Portuguese)',
        'Lucas Oliveira (Brazilian Portuguese)',
        'Lucas Loura (Brazilian Portuguese)',
        'Beruto666 (Brazilian Portuguese)',
        'twlvnn (Bulgarian)',
        'Roger VC (Catalan)',
        'Flynn (Cornish)'
        'Lukáš Linhart (Czech)',
        'Jakub Soukup (Czech)',
        'Petr Horník (Czech)',
        'Dingzhong Chen (Simplified Chinese)',
        'Eric-Song-Nop (Simplified Chinese)',
        'Inaha (Simplified Chinese)',
        'LS-Shandong (Simplified Chinese)',
        'randint (Traditional Chinese)',
        'Zhao Se (Traditional Chinese)',
        'happylittle7 (Traditional Chinese)',
        'Heimen Stoffels (Dutch)',
        'Philip Goto (Dutch)',
        'Koen Benne (Dutch)',
        'Mikachu (Dutch)',
        'Danial Behzadi (Persian)',
        'Muhammad Hussein Ammari (Persian)',
        'Jiri Grönroos (Finnish)',
        'Ricky Tigg (Finnish)',
        'Irénée THIRION (French)',
        'Valéry Febvre (French)',
        'Mathieu B. (French)',
        'rene-coty (French)',
        'paul verot (French)',
        'Temuri Doghonadze (Georgian)',
        'David Gogniashvili (Georgian)',
        'Sandor Odor (German)',
        'Liliana Prikler (German)',
        'gregorni (German)',
        'Liliana Marie Prikler (German)',
        'Tim (German)',
        'Sear Gasor (German)',
        'Vortex Acherontic (German)',
        'Dlurak (German)',
        'Mirko P. (German)',
        'Simon Barth (German)',
        'anon (German)',
        'Scrambled777 (Hindi)',
        'Milo Ivir (Croatian)',
        'Alifiyan Rosyidi (Indonesian)',
        'Alim Satria (Indonesian)',
        'Juan Manuel (Indonesian)',
        'srntskl-111 (Indonesian)',
        'Nataniel Dika Kurniawan (Indonesian)',
        'Mek101 (Italian)',
        'dedocc (Italian)',
        'Davide Mora (Italian)',
        'Andrea Scarano (Italian)',
        'pasquale ruotolo (Italian)',
        'Riccardo Luise (Italian)',
        'cas9 (Italian)',
        'Giulia (Italian)',
        'on9686 (Korean)',
        'Velyvis (Lithuanian)',
        'Lili Kurek (Polish)',
        'Aleksander Warzyniak (Polish)',
        'Kurai (Polish)',
        'Goraj (Polish)',
        'ssantos (Portuguese)',
        'Ademario Cunha (Portuguese)',
        'SpiralPack 527 (Portuguese)',
        'Lucas Silva Goulart (Portuguese)',
        'Manuela Silva (Portuguese)',
        'shima (Russian)',
        'Valentin Chernetsov (Russian)',
        'FIONover (Russian)',
        'Анна Алешкина #нетвойне (Russian)',
        'Сергей (Russian)',
        'Andrei Stepanov (Russian)',
        'Óscar Fernández Díaz (Spanish)',
        'gallegonovato (Spanish)',
        'Klauss (Spanish)',
        'Champiñon Traductor (Spanish)',
        'Libre (Spanish)',
        'Jesper (Swedish)',
        'PaneradFisk (Swedish)',
        'Willem Dinkelspiel (Swedish)',
        'Daniel Wiik (Swedish)',
        'தமிழ்நேரம் (Tamil)',
        'Ege Çelikçi (Turkish)',
        'Sabri Ünal (Turkish)',
        'Volkan Yıldırım (Turkish)',
        'Efe Akın (Turkish)',
        'Ahmet (Turkish)',
        'CakesTwix (Ukrainian)',
        'Kislotniy (Acela) (Ukrainian)',
        'mondstern (Ukrainian)',
        'DXCVII (Ukrainian)',
        'Bezruchenko Simon (Ukrainian)',
        'Максим Горпиніч (Ukrainian)',
        'Димко (Ukrainian)',
        'niyaki hayyashi (Vietnamese)',
        'Loc Huynh (Vietnamese)',
    ),
    supporters=(
        'gondolyr',
        'José',
    ),
)

RELEASE_NOTES = """
<p>This is a bugfix version.</p>
<ul>
    <li>[Reader] RTL/LTR/Vertical pager: Fixed keyboard/mouse navigation</li>
    <li>[Reader] OCR Translator: Improvements in Translator</li>
    <li>[L10n] Updated Russian and Ukrainian translations</li>
</ul>

<p>Changes in previous version 50.12.0</p>
<ul>
    <li>[Reader] Added an OCR translator</li>
    <li>[Reader] Moved settings from Menu to Settings dialog</li>
    <li>[Servers] Added MangaLivre (pt_BR)</li>
    <li>[Servers] Desu (RU): Update</li>
    <li>[Servers] JManga (JA): Update</li>
    <li>[Servers] MangaFire (EN/ES/FR/JA/PT/pt_BR): Update</li>
    <li>[Servers] Rncalation (ES): Update</li>
    <li>[Servers] Weeb Central (EN): Update</li>
    <li>[L10n] Updated Bengali, French, Korean, Polish, Russian and Ukrainian translations</li>
</ul>
<p>Happy reading.</p>
"""



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from urllib.parse import unquote_plus

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class Ninemanga(Server):
    id = 'ninemanga'
    name = 'Nine Manga'
    lang = 'en'
    is_nsfw = True
    status = 'disabled'  # Move to https://niadd.com

    base_url = 'https://www.ninemanga.com'
    logo_url = base_url + '/favicon.png'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/list/New-Update/'
    most_populars_url = base_url + '/list/Hot-Book/'
    manga_url = base_url + '/manga/{0}.html?waring=1'
    chapter_url = base_url + '/chapter/{0}/{1}-1.html'
    page_url = base_url + '/chapter/{0}/{1}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = {
                'User-Agent': USER_AGENT,
                'Accept-Language': 'fr-FR,fr;q=0.9,en;q=0.8,gl;q=0.7',
            }

    @classmethod
    def get_manga_initial_data_from_url(cls, url):
        return dict(slug=url.split('?')[0].split('/')[-1].replace('.html', ''))

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        if r.url == self.base_url:
            # Manga page doesn't exist, we have been redirected to homepage
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        name = soup.find('div', class_='ttline').h1.text.strip()
        name = name.replace(' Manga', '').replace(' Манга', '')  # cleaning
        data['name'] = name
        data['cover'] = soup.find('a', class_='bookface').img.get('src')

        # Details
        elements = soup.find('ul', class_='message').find_all('li')
        for element in elements:
            label = element.b.text

            if label.startswith(('Author', 'Auteur', 'Autor', 'Автор')):
                data['authors'] = [element.a.text.strip(), ]
            elif label.startswith(('Genre', 'Género', 'Genere', 'Gênero', 'Жанры')):
                for a_element in element.find_all('a'):
                    data['genres'].append(a_element.text)
            elif label.startswith(('Status', 'Statut', 'Estado', 'Stato', 'статус')):
                value = element.find_all('a')[0].text.strip().lower()

                if value in ('ongoing', 'en cours', 'laufende', 'en curso', 'in corso', 'em tradução', 'постоянный'):
                    data['status'] = 'ongoing'
                elif value in ('complete', 'completed', 'complété', 'abgeschlossen', 'completado', 'completato', 'completo', 'завершенный'):
                    data['status'] = 'complete'

        # Synopsis
        synopsis_element = soup.find('p', itemprop='description')
        if synopsis_element:
            synopsis_element.b.extract()
            data['synopsis'] = synopsis_element.text.strip()

        # Chapters
        if div_element := soup.find('div', class_='chapterbox'):
            li_elements = div_element.find_all('li')
            for li_element in reversed(li_elements):
                slug = li_element.a.get('href').split('/')[-1].replace('.html', '')
                data['chapters'].append(dict(
                    slug=slug,
                    title=li_element.a.text.strip(),
                    date=convert_date_string(li_element.span.text.strip(), format='%b %d, %Y'),
                ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')
        options_elements = soup.find('select', id='page').find_all('option')

        data = dict(
            pages=[],
        )
        for option_element in options_elements:
            data['pages'].append(dict(
                slug=option_element.get('value').split('/')[-1],
                image=None,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        # Scrap HTML page to get image url

        # Don't use `Referer` in headers, otherwise we are redirected to a page to select a mirror
        r = self.session_get(
            self.page_url.format(manga_slug, page['slug']),
            headers={
                'Accept':
                'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
                'Accept-Encoding': 'gzip, deflate, br',
                'Accept-Language': 'en-US,en;q=0.9,fr;q=0.8,fr-FR;q=0.7',
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')
        url = soup.select_one('#manga_pic_1').get('src')

        # Get scan image
        r = self.session_get(
            url,
            headers={
                'Referer': 'f{self.base_url}/',
            },
            timeout=30
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=url.split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns Latest Updates list
        """
        r = self.session_get(self.latest_updates_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.direlist .bookinfo'):
            a_element = element.select_one('.bookname')
            img_element = element.select_one('img')
            last_chapter_a_element = element.select_one('.chaptername')

            name = a_element.text.strip()
            results.append(dict(
                name=name,
                slug=unquote_plus(a_element.get('href')).split('/')[-1][:-5],
                cover=img_element.get('src'),
                last_chapter=last_chapter_a_element.text.replace(name, '').strip(),
            ))

        return results

    def get_most_populars(self):
        """
        Returns Hot manga list
        """
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.direlist .bookinfo'):
            a_element = element.select_one('.bookname')
            img_element = element.select_one('img')
            last_chapter_a_element = element.select_one('.chaptername')

            name = a_element.text.strip()
            results.append(dict(
                name=name,
                slug=unquote_plus(a_element.get('href')).split('/')[-1][:-5],
                cover=img_element.get('src'),
                last_chapter=last_chapter_a_element.text.replace(name, '').strip(),
            ))

        return results

    def search(self, term):
        r = self.session_get(
            self.search_url,
            params={
                'wd': term,
                # 'type': 'high',  # Advanced search
            },
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.direlist .bookinfo'):
            a_element = element.select_one('.bookname')
            img_element = element.select_one('img')
            last_chapter_a_element = element.select_one('.chaptername')

            name = a_element.text.strip()
            results.append(dict(
                name=name,
                slug=unquote_plus(a_element.get('href')).split('/')[-1][:-5],
                cover=img_element.get('src'),
                last_chapter=last_chapter_a_element.text.replace(name, '').strip(),
            ))

        return results


class Ninemanga_br(Ninemanga):
    # BEWARE: For historical reasons, Id is ninemanga_br instead of ninemanga_pt_br (idem for class name)
    id = 'ninemanga_br'
    lang = 'pt_BR'

    base_url = 'https://br.ninemanga.com'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/list/New-Update/'
    most_populars_url = base_url + '/list/Hot-Book/'
    manga_url = base_url + '/manga/{0}.html?waring=1'
    chapter_url = base_url + '/chapter/{0}/{1}-1.html'
    page_url = base_url + '/chapter/{0}/{1}'


class Ninemanga_de(Ninemanga):
    id = 'ninemanga_de'
    lang = 'de'

    base_url = 'https://de.ninemanga.com'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/list/New-Update/'
    most_populars_url = base_url + '/list/Hot-Book/'
    manga_url = base_url + '/manga/{0}.html?waring=1'
    chapter_url = base_url + '/chapter/{0}/{1}-1.html'
    page_url = base_url + '/chapter/{0}/{1}'


class Ninemanga_es(Ninemanga):
    id = 'ninemanga_es'
    lang = 'es'

    base_url = 'https://es.ninemanga.com'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/list/New-Update/'
    most_populars_url = base_url + '/list/Hot-Book/'
    manga_url = base_url + '/manga/{0}.html?waring=1'
    chapter_url = base_url + '/chapter/{0}/{1}-1.html'
    page_url = base_url + '/chapter/{0}/{1}'


class Ninemanga_fr(Ninemanga):
    id = 'ninemanga_fr'
    lang = 'fr'

    base_url = 'https://fr.ninemanga.com'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/list/New-Update/'
    most_populars_url = base_url + '/list/Hot-Book/'
    manga_url = base_url + '/manga/{0}.html?waring=1'
    chapter_url = base_url + '/chapter/{0}/{1}-1.html'
    page_url = base_url + '/chapter/{0}/{1}'


class Ninemanga_it(Ninemanga):
    id = 'ninemanga_it'
    lang = 'it'

    base_url = 'https://it.ninemanga.com'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/list/New-Update/'
    most_populars_url = base_url + '/list/Hot-Book/'
    manga_url = base_url + '/manga/{0}.html?waring=1'
    chapter_url = base_url + '/chapter/{0}/{1}-1.html'
    page_url = base_url + '/chapter/{0}/{1}'


class Ninemanga_ru(Ninemanga):
    id = 'ninemanga_ru'
    lang = 'ru'

    base_url = 'https://ru.ninemanga.com'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/list/New-Update/'
    most_populars_url = base_url + '/list/Hot-Book/'
    manga_url = base_url + '/manga/{0}.html?waring=1'
    chapter_url = base_url + '/chapter/{0}/{1}-1.html'
    page_url = base_url + '/chapter/{0}/{1}'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.manga_stream import MangaStream


class Rukavinari(MangaStream):
    id = 'rukavinari'
    name = 'Rukav Inari'
    lang = 'es'
    is_nsfw = True
    status = 'disabled'

    slug_position = -1

    base_url = 'https://rukavinari.org'

    authors_selector = '.tsinfo .imptdt:-soup-contains("Autor") i, .tsinfo .imptdt:-soup-contains("Artista") i'
    genres_selector = '.info-desc .mgen a'
    scanlators_selector = '.tsinfo .imptdt:-soup-contains("Serialización") i'
    status_selector = '.tsinfo .imptdt:-soup-contains("Estado") i'
    synopsis_selector = '[itemprop="description"]'
